"""Library for getting distance from optical sensors for the feedback.
Requires the nidaq module.
Assumes the following layout:

     (front)
  Ai0 ----- Ai3
     /     \
     |     |
 Ai1 \_ _ _/ Ai2

 Ai0 -> Front left
 Ai1 -> Back left
 Ai2 -> Back right
 Ai3 -> Front right

This library exports only one function: getLR()"""

import nidaq, numpy, math

def getLRT(win = False,medianFiltered = False):
    global sample
    read = nidaq.initNIDAQ()
    if(medianFiltered):
        samples = read(8)
        samples.sort(0)
        sample = samples[4]
    else:
        sample = read(1)[0]

    #Converts the first four voltages to distances; assumes that l1, l2, r1, r2 are
    #respectively connected to AI1 (Analog Input 1), AI2, AI0 and AI3 on the NIDAQ.
    l1,l2,r2,r1 = _VoltageToDistance(sample[:4])

    return _GetOffset(win,l1,l2,r1,r2)

def getLR(win = False,medianFiltered = False):
    l,r,t = getLRT(win,medianFiltered)
    return numpy.array([l,r])

def _VoltageToDistance(v):
    """Convert the voltage (V) at the leads of a Sharp GP2D12 IR Sensor to a distance in meters.
    Reliable only when the distance is between 0.01 and 0.08 meters. Works on numpy arrays as well as scalars."""

    return numpy.exp(numpy.log(142.1/v) / 0.88) / 1000.0

def rotate(x,y,theta):
    return [x*math.cos(theta)+y*math.sin(theta),-x*math.sin(theta)+y*math.cos(theta)]

def drawRotated(win,x0,y0,x1,y1,theta):
    [x0r,y0r] = rotate(x0,y0,theta)
    [x1r,y1r] = rotate(x1,y1,theta)
    win.drawLineSeg(x0r,y0r,x1r,y1r,'blue',2)

def oneside(win,r1,r2,alpha):
    x0 = -.03
    y0 = 0
    x1 = x0+(r1+.03)*math.cos(25*math.pi/180.)
    y1 = y0+(r1+.03)*math.sin(25*math.pi/180.)
    x2 = x0+(r2+.03)*math.cos(25*math.pi/180.)
    y2 = y0-(r2+.03)*math.sin(25*math.pi/180.)
    theta1 = -math.atan2(x1-x2,y1-y2)
    ll = .03+x1*math.cos(theta1)+y1*math.sin(theta1)
    if win==False:
        pass
    else:
        drawRotated(win,x0,y0,x1,y1,theta1+alpha)
        drawRotated(win,x0,y0,x2,y2,theta1+alpha)
        drawRotated(win,x1,y1,x2,y2,theta1+alpha)
        drawRotated(win,0,-.3,0,.3,theta1)
        drawRotated(win,-.05,.2,0,.3,theta1)
        drawRotated(win,.05,.2,0,.3,theta1)
    return numpy.array([ll,-theta1])

def _GetOffset(win,l1, l2, r1, r2):
    if win==False:
        pass
    else:
        win.clear()
    l,theta1=oneside(win,l2,l1,math.pi)
    r,theta2=oneside(win,r1,r2,0)
    if win==False:
        pass
    else:
        win.drawPoint(.03-l,0,'red')
        win.drawPoint(-.03+r,0,'red')
    return numpy.array([l,r,(theta1+theta2)/2.0])
