/*
    NIDAQserver is a server for National Instruments DAQ devices.
    Copyright (C) 2007  Karim Liman-Tinguiri <klt at mit.edu>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <ctype.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <wait.h>
#include "NIDAQmxBase.h"
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <getopt.h>
#include <math.h>

#define DAQmxErrChk(functionCall) { if( DAQmxFailed(error=(functionCall)) ) { goto Error; } }
#define VERSION 0.3

static char usage[] = "Usage: NIDAQserver [options]\n\n\
Valid options:\n\
--clocksource X (or -c X): sets the source of clock (default = OnboardClock)\n\
--help (or -h): displays this message and exits\n\
--min X (or -m X): sets the minimum expected voltage to X (default = -10)\n\
--max X (or -x X): sets the maximum expected voltage to X (default = 10)\n\
--pointstoread X (or -p X): sets the number of samples to acquire every (1 / samplerate) to X\n\
--samplerate X (or -r X): sets the sample rate to X\n\
--samplesperchan X (or -n X): sets the number of samples per channel (default = 100)\n\
--timeout X (or -t X): sets the timeout to X\n";

static int chanNumber(char * chanString) // returns the number of channels or -1 if the string is malformed
{
	// A valid input could be "Dev1/ai0, Dev1/ai1, Dev1/ai2"

	for(; *chanString && isspace(*chanString); chanString++);	// Skip spaces

	int chanNum = 0;	// Number of channels

	while(*chanString)
	{
		if (strncmp("Dev", chanString, sizeof("Dev") - 1) != 0)
			return -1;

		chanString += sizeof("Dev") - 1;

		if(!isdigit(*chanString))
			return -1;

		while(isdigit(*chanString))
			chanString++;	// Skips all digits (at least one of them)

		if(*chanString++ != '/')
			return -1;

		if (strncmp("ai", chanString, sizeof("ai") - 1) != 0)
			return -1;

		chanString += sizeof("ai") - 1;

		if(!isdigit(*chanString))
			return -1;

		while(isdigit(*chanString))
			chanString++;	// Skips all digits (at least one of them)

		while(*chanString && (isspace(*chanString) || *chanString == ','))
			chanString++;

		chanNum++;
	}

	return chanNum;	
}
int main(int argc, char *argv[])
{
	// Attempt to delete /tmp/DAQserver directory if it exists:
	system("rm -rf /tmp/NIDAQserver");
	if(mkdir("/tmp/NIDAQserver/", S_IRWXU | S_IRWXO))
	{	
		printf("Could not create /tmp/DAQserver/ directory\n");
		printf(strerror(errno));
		return -1;
	}

	if(mkfifo("/tmp/NIDAQserver/binary", S_IFIFO | 0666))
	{
		printf("Pipe creation failed\n");
		printf(strerror(errno));
		return -1;
	}
	
	int binaryPipe = 0;
	char pipeBuffer[4096] = {'\0'};

	// Default values below:
	
	// Task parameters
	int32	    error = 0;
	TaskHandle  taskHandle = 0;
	TaskHandle  taskHandle2 = 0; //output
	char        errBuff[2048]={'\0'};
	int32       i,j;
	time_t      startTime;
	bool32      done=0;

	// Channel parameters
	//char	*chan = "Dev1/ai0, Dev1/ai1, Dev1/ai2, Dev1/ai3, Dev1/ai4, Dev1/ai5, Dev1/ai6, Dev1/ai7";
	char	*chan = "Dev1/ai0, Dev1/ai1, Dev1/ai2, Dev1/ai3, Dev1/ai4, Dev1/ai5, Dev1/ai6, Dev1/ai7, Dev1/ai8, Dev1/ai9, Dev1/ai10, Dev1/ai11, Dev1/ai12, Dev1/ai13, Dev1/ai14, Dev1/ai15";
	int	numOfChan = chanNumber(chan);
	float64	min = -10.0;
	float64	max = 10.0;

	// Timing parameters
	char*	clockSource = "OnboardClock";
	uInt64	samplesPerChan = 8;
	float64	sampleRate = 200.0;

	// Data read parameters
	int32	pointsRead = 0;
	float64	timeout = 1.0;
	int32	totalRead = 0;
	
	// Output or not
	int output = 0;

	// Updates the values based on command-line arguments
	int o;

	while(1)
	{
		static struct option long_options[] =
		{
			{"min", required_argument, 0, 'm'},
			{"max", required_argument, 0, 'x'},
			{"chan", required_argument, 0, 'c'},
			{"clocksource", required_argument, 0, 's'},
			{"samplesperchan", required_argument, 0, 'n'},
			{"samplerate", required_argument, 0, 'r'},
			{"timeout", required_argument, 0, 't'},
			{"help", no_argument, 0, 'h'},
			{"output", no_argument, 0, 'o'},
			{0, 0, 0, 0}
		};

		int option_index = 0;

		o = getopt_long(argc, argv, "mxcsnrth", long_options, &option_index);
		
		if(o == -1)
			break;

		switch(o)
		{
			case 'c':
				chan = strcpy((char*)malloc(strlen(optarg) + 1),optarg);
				if((numOfChan = chanNumber(chan)) == -1)
				{
					printf("Could not parse channels. Got: %s\n", chan);
					return -1;	// Returning before calling any nidaqmxbase function
							// causes a segfault.
				}
				break;
			case 'm':
				min = atof(optarg);
				break;
			case 'x':
				max = atof(optarg);
				break;
			case 's':
				clockSource = optarg;
				break;
			case 'n':
				samplesPerChan = atoi(optarg);
				break;
			case 'r':
				sampleRate = atof(optarg);
				break;
			case 't':
				timeout = atof(optarg);
				break;
			case 'h':
				printf("%s", usage);
				return 0;
				break;
			case 'o':
                 		output = 1;
                 		break;
		}
	}
	
	if(output && mkfifo("/tmp/NIDAQserver/toNIDAQ", S_IFIFO | 0666))
	{
		printf("Pipe creation failed\n");
		printf(strerror(errno));
		return -1;
	}
	
	int toNIDAQpipe = 0;
	float64 prevOutput = 0;

	// Update data read parameters
	int     bufferSize = (uInt32)samplesPerChan * numOfChan;
	float64 *data = malloc(bufferSize * sizeof(float64));
	#define LENTOWRITE 128
	float64 *toWrite = malloc(LENTOWRITE * sizeof(float64));
	int32	pointsToRead = bufferSize;

	printf("NIDAQ server version %3.1f starting.\n"
		"Settings used are:\n"
		"Channels: %s\n"
		"Samples per Channel:\t%d\n"
		"Sample rate:\t\t%f\n"
		"Timeout:\t\t%f\n"
		"Minimum:\t\t%f\n"
		"Maximum:\t\t%f\n"
		"Clock source:\t\t%s\n", VERSION, chan, (int)samplesPerChan, (float)sampleRate,
			(float)timeout, (float)min, (float)max, clockSource);

	int logFile = open("/tmp/NIDAQserver/log", O_WRONLY | O_CREAT);
	
	if(logFile != (-1))
	{
		char logFileFormattedSettings[1024] = {'\0'};
		sprintf(logFileFormattedSettings,
			"ver=%f\n"
			"samplesperchan=%d\n"
			"samplerate=%d\n"
			"timeout=%f\n"
			"minimum=%f\n"
			"maximum=%f\n"
			"clocksource=\"%s\"\n"
			"channels=\"%s\"\n", VERSION, (int)samplesPerChan, (int)sampleRate,
			(float)timeout, (float)min, (float)max, clockSource, chan);
		write(logFile, logFileFormattedSettings, strlen(logFileFormattedSettings));
		close(logFile);
	}

	else
		printf("Failed to create /tmp/NIDAQserver/log\n");
		
	
	DAQmxErrChk (DAQmxBaseCreateTask("",&taskHandle));
	DAQmxErrChk (DAQmxBaseCreateAIVoltageChan(taskHandle,chan,"",DAQmx_Val_RSE,min,max,DAQmx_Val_Volts,NULL));
	if(output)
	{;
	DAQmxErrChk (DAQmxBaseCreateTask("",&taskHandle2));
	DAQmxErrChk (DAQmxBaseCreateAOVoltageChan(taskHandle2,"Dev1/ao0",NULL,0,5,DAQmx_Val_Volts,NULL));
	DAQmxErrChk (DAQmxBaseStartTask(taskHandle2));
	}
	DAQmxErrChk (DAQmxBaseCfgSampClkTiming(taskHandle,clockSource,sampleRate,DAQmx_Val_Rising,DAQmx_Val_ContSamps,samplesPerChan));
	DAQmxErrChk (DAQmxBaseStartTask(taskHandle));
	


	while( 1 )
	{

	        DAQmxErrChk (DAQmxBaseReadAnalogF64(taskHandle,
			samplesPerChan,timeout,DAQmx_Val_GroupByScanNumber,data,bufferSize,&pointsRead,NULL));
	        totalRead += pointsRead;
	        printf("Acquired %d samples. Total %d\n",pointsRead,totalRead);

		binaryPipe = open("/tmp/NIDAQserver/binary", O_WRONLY | O_NONBLOCK, S_IWRITE | S_IROTH);
	
		if(binaryPipe != (-1) && pointsRead)
		{
			int bytesWritten = write(binaryPipe, data, bufferSize * sizeof(float64) * numOfChan);

			if(bytesWritten > 0)
				printf("Wrote %d bytes to the binary pipe.\n", bytesWritten);
			else
				printf("Binary pipe error: %s\n", strerror(errno));

		}

		close(binaryPipe);
		
        if(output)
        {
                  toNIDAQpipe = open("/tmp/NIDAQserver/toNIDAQ", O_RDONLY | O_NONBLOCK);
                  if(toNIDAQpipe != (-1))
                  {

                           int bytesRead = read(toNIDAQpipe, toWrite, LENTOWRITE * sizeof(float64));
			   if(bytesRead < 0)
			   {
				        close(toNIDAQpipe);
				                  toNIDAQpipe = open("/tmp/NIDAQserver/toNIDAQ", O_RDONLY);
					bytesRead = read(toNIDAQpipe, toWrite, LENTOWRITE * sizeof(float64));
				// Someone attempted to write, retry the read blocking this time
			   }
                           if(bytesRead > 0)
                           {
                                        prevOutput = toWrite[(bytesRead / sizeof(float64)) - 1];
                           }
                  }
		  int32 written = 0;
                  DAQmxErrChk (DAQmxBaseWriteAnalogF64(taskHandle2, 1, FALSE, 1, DAQmx_Val_GroupByChannel, &prevOutput, &written, NULL));
		  printf("Wrote: %f to NIDAQ\n",prevOutput);

        }

        close(toNIDAQpipe);

	}

	close(binaryPipe);

Error:
	system("rm -rf /tmp/NIDAQserver");
	if( DAQmxFailed(error) )
		DAQmxBaseGetExtendedErrorInfo(errBuff,2048);
	if(taskHandle != 0)
	{
		DAQmxBaseStopTask (taskHandle);
		DAQmxBaseClearTask (taskHandle);
	}

	if( DAQmxFailed(error) )
	{
		printf("DAQmxBase Error: %s\n",errBuff);
	} 
	if(binaryPipe)
		close(binaryPipe);
	return 0;
}
