"""DAQ (Data AcQuisition) module for NI DAQ USB devices. Requires binary NIDAQserver."""

class NIDAQException (Exception):
    """Parent class for all exceptions thrown by the NIDAQ module."""
    def __str__(self):
        return self.__doc__

class NoNIDAQPresent(NIDAQException):
    """No NIDAQ is connected to the system."""

class NIDAQserverNotFound (NIDAQException):
    """Indicates a NIDAQserver binary could not be found in the current PATH nor in the current folder."""
    pass

class NIDAQServerException(NIDAQException):
    """Indicates the NIDAQserver was found but could not be used."""
    def __init__(self,error=""):
        self.error = error

    def __str__(self):
        return self.error

def initNIDAQ():
    """Initializes the NIDAQ and returns a procedure that can be used to acquire samples from it."""
    import os, time, numpy, sys

    NIDAQserverfile = None

    isNIDAQconnected = False

    # Look for NIDAQserver binary
    for path in os.environ.get('PATH').split(':') + ['.']:
        if os.path.exists(os.path.join(path,'NIDAQserver')):
            NIDAQserverfile = os.path.join(path,'NIDAQserver')
            break

    # Check for running NIDAQserver process
    isNIDAQRunning = lambda : os.popen('ps -A').read().find('NIDAQserver') != -1

    if NIDAQserverfile is None and not isNIDAQRunning(): #if neither exists, throw exception
        raise NIDAQserverNotFound()

    if not isNIDAQRunning(): #if NIDAQserver is not running, start it with default settings
        for line in os.popen('/etc/natinst/nidaqmxbase/bin/daqmxbase_listdevices').read().split('\n'):
            if line.find('Dev') != -1:
                isNIDAQconnected = True
                break

        if not isNIDAQconnected:
            raise NoNIDAQPresent()
        
        pipe = os.popen(NIDAQserverfile + " &")
	sys.stdout.write("Starting NIDAQserver, this may take a little while...")
	pipe.read(400)
	sys.stdout.write('done\n')	
        if not isNIDAQRunning():
            raise NIDAQServerException("Could not start the NIDAQserver")

    settings = {}
    exec open('/tmp/NIDAQserver/log') in settings
    numberOfChannels = len(settings['channels'].split(','))
    samplesPerChannel = int(settings['samplesperchan'])

    def read(samples=1):
        """Return a channels * /samples/ numpy array containing the /samples/ latest readings from the NIDAQ
        (the most recent reading is at index 0). For example, to access the 5th latest value of the third analog
        channel (Ai2), type: read(5)[4][2]. For most applications, pulling a single sample is sufficient."""

        try:
            binaryStream = open("/tmp/NIDAQserver/binary", "rb")
            chunk = binaryStream.read(samplesPerChannel * numberOfChannels * 8) # /numberOfChannels/ channels of 64 bits (8 bytes) doubles
            binaryStream.close() # We open and close the FIFO pipe each time to force the kernel to flush it, ensuring we have the latest data
        except IOException:
            raise NIDAQServerException('Could not acquire samples, make sure NIDAQ is plugged and call NIDAQ.initNIDAQ() again.')
        array = numpy.fromstring(chunk, numpy.float64, samplesPerChannel * numberOfChannels)
        array.shape = (array.size / numberOfChannels, numberOfChannels)
        return array[:samples]

    return read
